@@ -1,5 +1,5 @@
 # -*- coding: UTF-8 -*-
-#   Copyright 2012-2020 Fumail Project
+#   Copyright 2012-2022 Fumail Project
 #
 # Licensed under the Apache License, Version 2.0 (the "License");
 # you may not use this file except in compliance with the License.
@@ -16,6 +16,26 @@
 #
 #
 
+import logging
+
+class FuLoggerAdapter(logging.LoggerAdapter):
+    prefix = ''
+    
+    def __init__(self, logger, prefix=''):
+        super().__init__(logger, None)
+        self.prefix = prefix
+
+    def process(self, msg, kwargs):
+        if self.prefix:
+            return f"{self.prefix} {msg}", kwargs
+        else:
+            return msg, kwargs
+
+
+def get_logger(name, prefix=''):
+    return FuLoggerAdapter(logging.getLogger(name), prefix)
+
+
 ######################
 # new get_tld functions with support for subsubdomains
 ######################
@@ -1952,7 +1972,7 @@ confusables = {
 }
 
 
-def unconfuse(value):
+def unconfuse(value:str) -> str:
     try:
         string = ''
         for i in range(len(value)):
@@ -1962,7 +1982,7 @@ def unconfuse(value):
     return string
 
 
-def is_ascii(value):
+def is_ascii(value:str) -> bool:
     asciionly = True
     try:
         # only unconfuse if non-ascii characters are found
